import abc

import imgviz
import numpy as np
import numpy.typing as npt
from loguru import logger

from osam import types

from . import _decoding
from . import _encoding


class SamBase(types.Model):
    @abc.abstractmethod
    def _generate_mask_from_image_embedding(
        self, image_embedding: types.ImageEmbedding, prompt: types.Prompt
    ) -> npt.NDArray[np.bool_]:
        pass

    def generate(self, request: types.GenerateRequest) -> types.GenerateResponse:
        image_embedding: types.ImageEmbedding
        if request.image_embedding is None:
            if request.image is None:
                raise ValueError("request.image or request.image_embedding is required")
            image_embedding = self.encode_image(request.image)
        else:
            image_embedding = request.image_embedding

        prompt: types.Prompt
        if request.prompt is None:
            prompt = types.Prompt(
                points=np.array(
                    [
                        [
                            image_embedding.original_width / 2,
                            image_embedding.original_height / 2,
                        ]
                    ],
                    dtype=np.float32,
                ),
                point_labels=np.array([1], dtype=np.int32),
            )
            logger.warning(
                "Prompt is not given, so using the center point as prompt: {prompt!r}",
                prompt=prompt,
            )
        else:
            prompt = request.prompt
        del request

        if prompt.points is None or prompt.point_labels is None:
            raise ValueError("Prompt must contain points and point_labels: %r", prompt)

        mask: npt.NDArray[np.bool_] = self._generate_mask_from_image_embedding(
            image_embedding=image_embedding, prompt=prompt
        )

        bbox = imgviz.masks_to_bboxes(masks=[mask])[0].astype(int)
        bounding_box: types.BoundingBox = types.BoundingBox(
            ymin=bbox[0], xmin=bbox[1], ymax=bbox[2], xmax=bbox[3]
        )

        return types.GenerateResponse(
            model=self.name,
            image_embedding=image_embedding,
            annotations=[types.Annotation(mask=mask, bounding_box=bounding_box)],
        )


class Sam(SamBase):
    def encode_image(self, image: np.ndarray) -> types.ImageEmbedding:
        return _encoding.compute_image_embedding_from_image(
            encoder_session=self._inference_sessions["encoder"],
            image=image,
        )

    def _generate_mask_from_image_embedding(
        self, image_embedding: types.ImageEmbedding, prompt: types.Prompt
    ) -> npt.NDArray[np.bool_]:
        return _decoding.generate_mask_from_image_embedding(
            decoder_session=self._inference_sessions["decoder"],
            image_embedding=image_embedding,
            prompt=prompt,
            input_size=_encoding.get_input_size(
                encoder_session=self._inference_sessions["encoder"]
            ),
        )


class Sam100m(Sam):
    name = "sam:100m"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/wkentaro/labelme/releases/download/sam-20230416/sam_vit_b_01ec64.quantized.encoder.onnx",
            hash="sha256:3346b9cc551c9902fbf3b203935e933592b22e042365f58321c17fc12641fd6a",
        ),
        "decoder": types.Blob(
            url="https://github.com/wkentaro/labelme/releases/download/sam-20230416/sam_vit_b_01ec64.quantized.decoder.onnx",
            hash="sha256:edbcf1a0afaa55621fb0abe6b3db1516818b609ea9368f309746a3afc68f7613",
        ),
    }


class Sam300m(Sam):
    name = "sam:300m"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/wkentaro/labelme/releases/download/sam-20230416/sam_vit_l_0b3195.quantized.encoder.onnx",
            hash="sha256:f7158a4a1fe7f670ef47ea2f7f852685425c1ed6caa40f5df86cbe2b0502034f",
        ),
        "decoder": types.Blob(
            url="https://github.com/wkentaro/labelme/releases/download/sam-20230416/sam_vit_l_0b3195.quantized.decoder.onnx",
            hash="sha256:552ebb23bf52c5e5b971ac710d1eb8dccfd88b36cc6aff881d1536d1662e6d7b",
        ),
    }


class Sam600m(Sam):
    name = "sam:latest"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/wkentaro/labelme/releases/download/sam-20230416/sam_vit_h_4b8939.quantized.encoder.onnx",
            hash="sha256:a5c745fd4279efc5e5436b412200e983dafc2249ce172af6cc0002a71bb5f485",
        ),
        "decoder": types.Blob(
            url="https://github.com/wkentaro/labelme/releases/download/sam-20230416/sam_vit_h_4b8939.quantized.decoder.onnx",
            hash="sha256:020b385a45ffe51097e1acd10cd791075a86171153505f789a793bc382eef210",
        ),
    }



from typing import Final

from loguru import logger

import osam


def pytest_configure(config):
    config.addinivalue_line("markers", "extra: marks tests as extra")

    if config.option.collectonly:
        return
    if _is_slave := hasattr(config, "workerinput"):  # noqa: F841
        return

    default_test_models_only: bool = "not extra" in config.option.markexpr
    DEFAULT_TEST_MODELS: Final[list[str]] = [
        "efficientsam:10m",
        "sam3:latest",
        "sam2:tiny",
    ]

    # download should not happen in parallel
    logger.info("Pulling all registered model types")
    for model in osam.apis.registered_model_types:
        if default_test_models_only and model not in DEFAULT_TEST_MODELS:
            continue
        model.pull()
    logger.info("Finished pulling all registered model types")



#!/usr/bin/env python

import argparse
import time

import numpy as np
import PIL.Image
from loguru import logger

import osam.apis
import osam.types


def benchmark(model: str, n_times: int):
    request = osam.types.GenerateRequest(
        model=model,
        image=np.asarray(PIL.Image.open("../_images/dogs.jpg")),
        prompt=osam.types.Prompt(
            points=np.array([[1280, 800]]), point_labels=np.array([1])
        ),
    )

    logger.info("Warming up")
    for _ in range(3):
        osam.apis.generate(request)
    logger.info("Finished warming up")

    logger.info(f"Benchmarking {n_times} times")
    elapsed_times = []
    for _ in range(n_times):
        t_start = time.time()
        osam.apis.generate(request)
        elapsed_time = time.time() - t_start
        elapsed_times.append(elapsed_time)
    logger.info(f"Average elapsed time: {np.mean(elapsed_times)} [s]")


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "model",
        type=str,
        choices=[model_type.name for model_type in osam.apis.registered_model_types],
    )
    args = parser.parse_args()

    benchmark(model=args.model, n_times=10)


if __name__ == "__main__":
    main()



import collections


def tabulate(rows, headers):
    col_widths = collections.defaultdict(int)
    for row in [headers] + rows:
        for c, item in enumerate(row):
            col_widths[c] = max(col_widths[c], len(item))

    table = ""

    for c, header in enumerate(headers):
        table += header.ljust(col_widths[c] + 4)

    for row in rows:
        table += "\n"
        for c, item in enumerate(row):
            table += item.ljust(col_widths[c] + 4)
    return table



import numpy as np
import numpy.typing as npt

from osam import types
from osam._models.sam import SamBase

from . import _decoding
from . import _encoding


class EfficientSam(SamBase):
    def encode_image(self, image: np.ndarray) -> types.ImageEmbedding:
        return _encoding.compute_image_embedding_from_image(
            encoder_session=self._inference_sessions["encoder"],
            image=image,
        )

    def _generate_mask_from_image_embedding(
        self,
        image_embedding: types.ImageEmbedding,
        prompt: types.Prompt,
    ) -> npt.NDArray[np.bool_]:
        return _decoding.generate_mask_from_image_embedding(
            decoder_session=self._inference_sessions["decoder"],
            image_embedding=image_embedding,
            prompt=prompt,
        )


class EfficientSam10m(EfficientSam):
    name = "efficientsam:10m"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/labelmeai/efficient-sam/releases/download/onnx-models-20231225/efficient_sam_vitt_encoder.onnx",
            hash="sha256:7a73ee65aa2c37237c89b4b18e73082f757ffb173899609c5d97a2bbd4ebb02d",
        ),
        "decoder": types.Blob(
            url="https://github.com/labelmeai/efficient-sam/releases/download/onnx-models-20231225/efficient_sam_vitt_decoder.onnx",
            hash="sha256:e1afe46232c3bfa3470a6a81c7d3181836a94ea89528aff4e0f2d2c611989efd",
        ),
    }


class EfficientSam30m(EfficientSam):
    name = "efficientsam:latest"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/labelmeai/efficient-sam/releases/download/onnx-models-20231225/efficient_sam_vits_encoder.onnx",
            hash="sha256:4cacbb23c6903b1acf87f1d77ed806b840800c5fcd4ac8f650cbffed474b8896",
        ),
        "decoder": types.Blob(
            url="https://github.com/labelmeai/efficient-sam/releases/download/onnx-models-20231225/efficient_sam_vits_decoder.onnx",
            hash="sha256:4727baf23dacfb51d4c16795b2ac382c403505556d0284e84c6ff3d4e8e36f22",
        ),
    }



from typing import cast

import numpy as np
import numpy.typing as npt
import onnxruntime

from osam import types


def generate_mask_from_image_embedding(
    decoder_session: onnxruntime.InferenceSession,
    image_embedding: types.ImageEmbedding,
    prompt: types.Prompt,
    input_size: int,
) -> npt.NDArray[np.bool_]:
    input_point: npt.NDArray[np.float32] = np.array(prompt.points, dtype=np.float32)
    input_point = input_point / np.array(
        [
            image_embedding.original_width / input_size,
            image_embedding.original_height / input_size,
        ],
        dtype=np.float32,
    )
    input_label: npt.NDArray[np.float32] = np.array(
        prompt.point_labels, dtype=np.float32
    )

    decoder_inputs = {
        "image_embeddings": image_embedding.embedding[None],
        "high_res_features1": image_embedding.extra_features[0][None],
        "high_res_features2": image_embedding.extra_features[1][None],
        "point_coords": input_point[None],
        "point_labels": input_label[None],
        "mask_input": np.zeros((1, 1, 256, 256), dtype=np.float32),
        "has_mask_input": np.array([0], dtype=np.float32),
        "orig_im_size": np.array(
            (image_embedding.original_height, image_embedding.original_width),
            dtype=np.int64,
        ),
    }
    masks, scores, _low_res_mask = decoder_session.run(None, decoder_inputs)
    masks = cast(npt.NDArray[np.float32], masks)
    scores = cast(npt.NDArray[np.float32], scores)

    mask: npt.NDArray[np.bool_] = (
        masks[0, np.argmax(scores)] > 0.0
    )  # (1, N, H, W) -> (H, W)
    return mask



from typing import Optional
from typing import Type
from typing import cast

import numpy as np
import onnxruntime
from numpy.typing import NDArray

from . import _models
from . import types

running_model: Optional[types.Model] = None

registered_model_types: list[Type[types.Model]] = [
    _models.EfficientSam10m,
    _models.EfficientSam30m,
    _models.Sam100m,
    _models.Sam2BasePlus,
    _models.Sam2Large,
    _models.Sam2Small,
    _models.Sam2Tiny,
    _models.Sam3,
    _models.Sam300m,
    _models.Sam600m,
    _models.YoloWorldXL,
]


def get_model_type_by_name(name: str) -> Type[types.Model]:
    model_name: str
    if ":" in name:
        model_name = name
    else:
        model_name = f"{name}:latest"

    for cls in registered_model_types:
        if cls.name == model_name:
            break
    else:
        raise ValueError(f"Model {name!r} not found.")
    return cls


def generate(request: types.GenerateRequest) -> types.GenerateResponse:
    global running_model

    model_type = get_model_type_by_name(name=request.model)
    if running_model is None or running_model.name != model_type.name:
        running_model = model_type()
    assert running_model is not None

    response: types.GenerateResponse = running_model.generate(request=request)
    return response


_non_maximum_suppression_inference_session: Optional[
    onnxruntime.InferenceSession
] = None


def non_maximum_suppression(
    boxes: NDArray[np.floating],
    scores: NDArray[np.floating],
    iou_threshold: float,
    score_threshold: float,
    max_num_detections: int,
) -> tuple[
    NDArray[np.floating], NDArray[np.floating], NDArray[np.int64], NDArray[np.int64]
]:
    global _non_maximum_suppression_inference_session
    if _non_maximum_suppression_inference_session is None:
        blob = types.Blob(
            url="https://github.com/wkentaro/yolo-world-onnx/releases/download/v0.1.0/non_maximum_suppression.onnx",  # noqa
            hash="sha256:328310ba8fdd386c7ca63fc9df3963cc47b1268909647abd469e8ebdf7f3d20a",
        )
        blob.pull()
        _non_maximum_suppression_inference_session = onnxruntime.InferenceSession(
            blob.path, providers=["CPUExecutionProvider"]
        )
    inference_session = _non_maximum_suppression_inference_session

    outputs = inference_session.run(
        output_names=["selected_indices"],
        input_feed={
            "boxes": boxes[None, :, :],
            "scores": scores[None, :, :].transpose(0, 2, 1),
            "max_output_boxes_per_class": np.array(
                [max_num_detections], dtype=np.int64
            ),
            "iou_threshold": np.array([iou_threshold], dtype=np.float32),
            "score_threshold": np.array([score_threshold], dtype=np.float32),
        },
    )
    selected_indices = cast(NDArray[np.int64], outputs[0])
    labels = selected_indices[:, 1]
    box_indices = selected_indices[:, 2]
    boxes = boxes[box_indices]
    scores = scores[box_indices, labels]
    indices = box_indices

    if len(boxes) > max_num_detections:
        keep_indices = np.argsort(scores)[-max_num_detections:]
        boxes = boxes[keep_indices]
        scores = scores[keep_indices]
        labels = labels[keep_indices]
        indices = indices[keep_indices]

    return boxes, scores, labels, indices



from typing import cast

import numpy as np
import numpy.typing as npt
import onnxruntime

from osam import types


def compute_image_embedding_from_image(
    encoder_session: onnxruntime.InferenceSession, image: npt.NDArray[np.uint8]
) -> types.ImageEmbedding:
    if image.ndim == 2:
        raise ValueError("Grayscale images are not supported")
    if image.ndim == 3 and image.shape[2] == 4:
        raise ValueError("RGBA images are not supported")

    batched_images: npt.NDArray[np.float32] = (
        image.transpose(2, 0, 1)[None].astype(np.float32) / 255
    )
    outputs = encoder_session.run(
        output_names=None,
        input_feed={"batched_images": batched_images},
    )
    image_embedding: npt.NDArray[np.float32] = cast(
        npt.NDArray[np.float32], outputs[0]
    )[0]  # (embedding_dim, height, width)

    return types.ImageEmbedding(
        original_height=image.shape[0],
        original_width=image.shape[1],
        embedding=image_embedding,
    )



from typing import cast

import numpy as np
import numpy.typing as npt
import onnxruntime

from osam import types

from . import _images


def get_input_size(encoder_session: onnxruntime.InferenceSession) -> int:
    input_height: int
    input_width: int
    input_height, input_width = encoder_session.get_inputs()[0].shape[2:]
    if input_height != input_width:
        raise ValueError("Input height and width must be equal")
    return input_height


def _compute_input_from_image(
    image: npt.NDArray[np.uint8], input_size: int
) -> npt.NDArray[np.float32]:
    _scale, scaled_image = _images.resize_image(image=image, target_size=input_size)
    input_: npt.NDArray[np.float32] = (
        scaled_image.astype(np.float32)
        - np.array([123.675, 116.28, 103.53], dtype=np.float32)
    ) / np.array([58.395, 57.12, 57.375], dtype=np.float32)
    input_ = np.pad(
        input_,
        (
            (0, input_size - input_.shape[0]),
            (0, input_size - input_.shape[1]),
            (0, 0),
        ),
    )
    input_ = input_.transpose(2, 0, 1)[None, :, :, :]
    return input_


def compute_image_embedding_from_image(
    encoder_session: onnxruntime.InferenceSession,
    image: npt.NDArray[np.uint8],
) -> types.ImageEmbedding:
    if image.ndim == 2:
        raise ValueError("Grayscale images are not supported")
    if image.ndim == 3 and image.shape[2] == 4:
        raise ValueError("RGBA images are not supported")

    input_: npt.NDArray[np.float32] = _compute_input_from_image(
        image=image, input_size=get_input_size(encoder_session=encoder_session)
    )
    outputs = encoder_session.run(output_names=None, input_feed={"x": input_})
    image_embedding: npt.NDArray[np.float32] = cast(
        npt.NDArray[np.float32], outputs[0]
    )[0]  # (embedding_dim, height, width)

    return types.ImageEmbedding(
        original_height=image.shape[0],
        original_width=image.shape[1],
        embedding=image_embedding,
    )



import datetime
import math


def naturalsize(size: int) -> str:
    """Convert byte size to human readable format."""
    if size == 0:
        return "0B"
    base = 1024
    prefixes = ["B", "KB", "MB", "GB", "TB", "PB"]
    power = int(math.log(size, base))
    prefix = prefixes[power]
    size /= base**power
    return f"{size:.2f} {prefix}"


def naturaltime(dt: datetime.datetime) -> str:
    """Convert datetime to human readable format."""
    now = datetime.datetime.now()
    delta = now - dt
    if delta.days > 0:
        return f"{delta.days} days ago"
    elif delta.seconds < 60:
        return f"{delta.seconds} seconds ago"
    elif delta.seconds < 3600:
        return f"{delta.seconds // 60} minutes ago"
    else:
        return f"{delta.seconds // 3600} hours ago"



from typing import cast

import imgviz
import numpy as np
import numpy.typing as npt
import onnxruntime

from osam import types


def get_input_size(encoder_session: onnxruntime.InferenceSession) -> int:
    input_height: int
    input_width: int
    input_height, input_width = encoder_session.get_inputs()[0].shape[2:]
    if input_height != input_width:
        raise ValueError("Input height and width must be equal")
    return input_height


def compute_image_embedding_from_image(
    encoder_session: onnxruntime.InferenceSession, image: npt.NDArray[np.uint8]
) -> types.ImageEmbedding:
    if image.ndim == 2:
        raise ValueError("Grayscale images are not supported")
    if image.ndim == 3 and image.shape[2] == 4:
        raise ValueError("RGBA images are not supported")

    input_height: int
    input_width: int
    input_height, input_width = encoder_session.get_inputs()[0].shape[2:]

    input_: npt.NDArray[np.float32]
    input_ = (
        imgviz.resize(image, width=input_width, height=input_height).astype(np.float32)
        / 255
    )
    #
    input_ = input_ - np.array([0.485, 0.456, 0.406], dtype=np.float32) / np.array(
        [0.229, 0.224, 0.225], dtype=np.float32
    )
    input_ = input_.transpose(2, 0, 1)[None]

    outputs = encoder_session.run(output_names=None, input_feed={"input": input_})
    image_embedding = cast(npt.NDArray[np.float32], outputs[0])
    high_res_features1 = cast(npt.NDArray[np.float32], outputs[1])
    high_res_features2 = cast(npt.NDArray[np.float32], outputs[2])

    return types.ImageEmbedding(
        original_height=image.shape[0],
        original_width=image.shape[1],
        embedding=image_embedding[0],
        extra_features=[high_res_features1[0], high_res_features2[0]],
    )



import io
import os
import shlex
import subprocess

import PIL.Image

here = os.path.dirname(os.path.abspath(__file__))


def test_run():
    cmd = f"osam run efficientsam:10m --image {here}/../examples/_images/dogs.jpg"
    output = subprocess.check_output(shlex.split(cmd))
    image = PIL.Image.open(io.BytesIO(output))
    assert image.size == (2560, 1600)



import numpy as np
import numpy.typing as npt

from osam import types
from osam._models.sam import SamBase

from . import _decoding
from . import _encoding


class Sam2(SamBase):
    def encode_image(self, image: npt.NDArray[np.uint8]) -> types.ImageEmbedding:
        return _encoding.compute_image_embedding_from_image(
            encoder_session=self._inference_sessions["encoder"], image=image
        )

    def _generate_mask_from_image_embedding(
        self,
        image_embedding: types.ImageEmbedding,
        prompt: types.Prompt,
    ) -> npt.NDArray[np.bool_]:
        return _decoding.generate_mask_from_image_embedding(
            decoder_session=self._inference_sessions["decoder"],
            image_embedding=image_embedding,
            prompt=prompt,
            input_size=_encoding.get_input_size(
                encoder_session=self._inference_sessions["encoder"]
            ),
        )


class Sam2Tiny(Sam2):
    name = "sam2:tiny"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/wkentaro/osam/releases/download/sam2.1/sam2.1_tiny_preprocess.onnx",
            hash="sha256:5557482c56565f6a6c8206874b1a11c392cef8a1766477bf035b919092f2b619",
        ),
        "decoder": types.Blob(
            url="https://github.com/wkentaro/osam/releases/download/sam2.1/sam2.1_tiny.onnx",
            hash="sha256:11a2c86fabbea9d0268213a9205c99a7f7e379caa0493bd13f5cca8ffaae6777",
        ),
    }


class Sam2Small(Sam2):
    name = "sam2:small"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/wkentaro/osam/releases/download/sam2.1/sam2.1_small_preprocess.onnx",
            hash="sha256:06016c6dfb146ce10e4dadfdf49e88a05c8d1f97a6b7c57e150e60d2d46a72e7",
        ),
        "decoder": types.Blob(
            url="https://github.com/wkentaro/osam/releases/download/sam2.1/sam2.1_small.onnx",
            hash="sha256:153aaef5047a3b95285d90cbb39dad6c7b5821bfd944dbf56483f3f735936941",
        ),
    }


class Sam2BasePlus(Sam2):
    name = "sam2:latest"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/wkentaro/osam/releases/download/sam2.1/sam2.1_base_plus_preprocess.onnx",
            hash="sha256:ce95c44082b4532c25ae01e11da3c9337dab7b04341455c09ae599dc9ae5c438",
        ),
        "decoder": types.Blob(
            url="https://github.com/wkentaro/osam/releases/download/sam2.1/sam2.1_base_plus.onnx",
            hash="sha256:2ad091af889b20ad2035503b4355cd8924fcf0e29fa6536924c48dc220ecdc56",
        ),
    }


class Sam2Large(Sam2):
    name = "sam2:large"

    _blobs = {
        "encoder": types.Blob(
            url="https://github.com/wkentaro/osam/releases/download/sam2.1/sam2.1_large_preprocess.onnx",
            hash="sha256:ab676f957528918496990f242163fd6b41a7222ae255862e846d9ab35115c12e",
        ),
        "decoder": types.Blob(
            url="https://github.com/wkentaro/osam/releases/download/sam2.1/sam2.1_large.onnx",
            hash="sha256:a3ebc6b8e254bd4ca1346901b9472bc2fae9e827cfd67d67e162d0ae2b1ec9a0",
        ),
    }



import datetime
import io
import json
import os
import sys

import click
import imgviz
import numpy as np
import PIL.Image
from loguru import logger
from numpy.typing import NDArray

from . import __version__
from . import _humanize
from . import _tabulate
from . import apis
from . import types


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(__version__)
def cli():
    logger.remove(0)
    logger.add(
        sys.stderr,
        level="INFO",
        colorize=True,
        format="<level>{message}</level>",
        backtrace=False,
        diagnose=False,
    )
    os.makedirs(os.path.expanduser("~/.cache/osam"), exist_ok=True)
    logger.add(
        os.path.expanduser("~/.cache/osam/osam.log"), colorize=True, level="DEBUG"
    )


@cli.command(help="Help about any command")
@click.argument("subcommand", required=False, type=str)
@click.pass_context
def help(ctx, subcommand):
    if subcommand is None:
        click.echo(cli.get_help(ctx))
        return

    subcommand_obj = cli.get_command(ctx, subcommand)
    if subcommand_obj is None:
        logger.warning("Unknown subcommand {subcommand!r}", subcommand=subcommand)
        click.echo(cli.get_help(ctx))
    else:
        click.echo(subcommand_obj.get_help(ctx))


@cli.command(help="List models")
@click.option("--all", "-a", "show_all", is_flag=True, help="show all models")
def list(show_all):
    rows = []
    for model_type in apis.registered_model_types:
        size = model_type.get_size()
        modified_at = model_type.get_modified_at()

        if size is None or modified_at is None:
            if show_all:
                size = "<not pulled>"
                modified_at = "<not pulled>"
            else:
                continue
        else:
            size = _humanize.naturalsize(size)
            modified_at = _humanize.naturaltime(
                datetime.datetime.fromtimestamp(modified_at)
            )

        rows.append([model_type.name, model_type.get_id(), size, modified_at])
    click.echo(_tabulate.tabulate(rows, headers=["NAME", "ID", "SIZE", "MODIFIED"]))


@cli.command(help="Pull a model")
@click.argument("model_name", metavar="model", type=str)
def pull(model_name):
    cls = apis.get_model_type_by_name(model_name)
    logger.info("Pulling {model_name!r}...", model_name=model_name)
    cls.pull()
    logger.info("Pulled {model_name!r}", model_name=model_name)


@cli.command(help="Remove a model")
@click.argument("model_name", metavar="model", type=str)
def rm(model_name):
    cls = apis.get_model_type_by_name(model_name)
    logger.info("Removing {model_name!r}...", model_name=model_name)
    cls.remove()
    logger.info("Removed {model_name!r}", model_name=model_name)


@cli.command(help="Start server")
@click.option("--reload", is_flag=True, help="reload server on file changes")
def serve(reload):
    try:
        import uvicorn  # type: ignore[unresolved-import]

        import osam._server  # noqa: F401

        logger.info("Starting server...")
        uvicorn.run("osam._server:app", host="127.0.0.1", port=11368, reload=reload)
    except ImportError:
        logger.error("Run `pip install osam[serve]` to use `osam serve`")
        sys.exit(1)


@cli.command(help="Run a model")
@click.argument("model_name", metavar="model", type=str)
@click.option(
    "--image",
    "image_path",
    type=click.Path(exists=True),
    help="image path",
    required=True,
)
@click.option("--prompt", type=json.loads, help="prompt")
@click.option("--json", is_flag=True, help="json output")
def run(model_name: str, image_path: str, prompt, json: bool) -> None:
    image: np.ndarray = np.asarray(PIL.Image.open(image_path))

    try:
        request: types.GenerateRequest = types.GenerateRequest(
            model=model_name, image=image, prompt=prompt
        )
        response: types.GenerateResponse = apis.generate(request=request)
    except Exception:
        logger.exception("Failed to run model")
        sys.exit(1)

    if json:
        click.echo(response.model_dump_json())
    else:
        visualization: np.ndarray

        if request.prompt and request.prompt.texts is not None:
            labels = [
                0
                if annotation.text is None
                else 1 + request.prompt.texts.index(annotation.text)
                for annotation in response.annotations
            ]
        else:
            labels = [1] * len(response.annotations)

        captions = []
        for annotation in response.annotations:
            if annotation.text is not None and annotation.score is not None:
                caption = f"{annotation.text}: {annotation.score:.2f}"
            elif annotation.text is not None:
                caption = f"{annotation.text}"
            elif annotation.score is not None:
                caption = f"{annotation.score:.2f}"
            else:
                caption = None
            captions.append(caption)

        bboxes: NDArray[np.float32] | None = None
        if all(
            annotation.bounding_box is not None for annotation in response.annotations
        ):
            bboxes = np.array(
                [
                    [
                        getattr(annotation.bounding_box, key)
                        for key in ["ymin", "xmin", "ymax", "xmax"]
                    ]
                    for annotation in response.annotations
                ],
                dtype=np.float32,
            )

        masks: NDArray[np.bool_] | None = None
        if all(annotation.mask is not None for annotation in response.annotations):
            masks = np.array([annotation.mask for annotation in response.annotations])

        visualization = imgviz.instances2rgb(
            image=image,
            labels=labels,
            bboxes=bboxes,
            masks=masks,
            captions=captions,
            alpha=0.5,
        )

        sys.stdout.buffer.write(_image_ndarray_to_data(visualization))


def _image_ndarray_to_data(ndarray: np.ndarray) -> bytes:
    pil = PIL.Image.fromarray(ndarray)
    with io.BytesIO() as f:
        pil.save(f, format="PNG")
        return f.getvalue()


if __name__ == "__main__":
    cli()



from typing import cast

import numpy as np
import PIL.Image
from loguru import logger
from numpy.typing import NDArray

from osam import types

from ..yoloworld.clip import tokenize


class Sam3(types.Model):
    name = "sam3:latest"

    _blobs = {
        "image_encoder": types.Blob(
            url="https://huggingface.co/wkentaro/sam3-onnx-models/resolve/main/sam3_image_encoder.onnx",
            hash="sha256:3dd676271e9c459463f7026d8ab2c6318672cd89511f89c30b34cdf0a2e67a3f",
            attachments=[
                types.Blob(
                    url="https://huggingface.co/wkentaro/sam3-onnx-models/resolve/main/sam3_image_encoder.onnx.data",
                    hash="sha256:03bd50b0703e2b04e2193ca831b7f9d5ecf40bc5287cc59b1970f56ab800c995",
                ),
            ],
        ),
        "language_encoder": types.Blob(
            url="https://huggingface.co/wkentaro/sam3-onnx-models/resolve/main/sam3_language_encoder.onnx",
            hash="sha256:af80dffcd8fc369281b0422295244ba22f64e8424e9eefa111ac8966e9ea9ba6",
            attachments=[
                types.Blob(
                    url="https://huggingface.co/wkentaro/sam3-onnx-models/resolve/main/sam3_language_encoder.onnx.data",
                    hash="sha256:1b03dabc657c1f2887d1b8ce2a3537467d4c033ea1f8c8be141fbca55e4b95f7",
                ),
            ],
        ),
        "decoder": types.Blob(
            url="https://huggingface.co/wkentaro/sam3-onnx-models/resolve/main/sam3_decoder.onnx",
            hash="sha256:aca3109a1bf87d1589bf2a4a61c641fd97624152b21c6e9a5aa85735db398884",
            attachments=[
                types.Blob(
                    url="https://huggingface.co/wkentaro/sam3-onnx-models/resolve/main/sam3_decoder.onnx.data",
                    hash="sha256:afae2f057f4c6e2478589877aa3d361a0d863cdfdb8259f7eee903f529188ac6",
                ),
            ],
        ),
    }

    def encode_image(self, image: NDArray[np.uint8]) -> types.ImageEmbedding:
        logger.debug("Encoding image with shape: {}", image.shape)
        original_height: int = image.shape[0]
        original_width: int = image.shape[1]

        pil_image: PIL.Image.Image = PIL.Image.fromarray(image).resize((1008, 1008))
        image_input: NDArray[np.uint8] = np.asarray(pil_image).transpose(
            2, 0, 1
        )  # HWC → CHW

        outputs = self._inference_sessions["image_encoder"].run(
            None, {"image": image_input}
        )

        vision_pos_enc_2: NDArray[np.float32] = cast(
            NDArray[np.float32], outputs[2]
        ).squeeze(axis=0)
        backbone_fpn_0: NDArray[np.float32] = cast(
            NDArray[np.float32], outputs[3]
        ).squeeze(axis=0)
        backbone_fpn_1: NDArray[np.float32] = cast(
            NDArray[np.float32], outputs[4]
        ).squeeze(axis=0)
        backbone_fpn_2: NDArray[np.float32] = cast(
            NDArray[np.float32], outputs[5]
        ).squeeze(axis=0)

        return types.ImageEmbedding(
            original_height=original_height,
            original_width=original_width,
            embedding=backbone_fpn_0,
            extra_features=[backbone_fpn_1, backbone_fpn_2, vision_pos_enc_2],
        )

    def generate(self, request: types.GenerateRequest) -> types.GenerateResponse:
        image_embedding: types.ImageEmbedding
        if request.image_embedding is not None:
            image_embedding = request.image_embedding
        elif request.image is not None:
            image_embedding = self.encode_image(image=request.image)
        else:
            raise ValueError("either image or image_embedding is required for Sam3")

        original_height: int = image_embedding.original_height
        original_width: int = image_embedding.original_width

        backbone_fpn_0: NDArray[np.float32] = image_embedding.embedding[None]
        backbone_fpn_1: NDArray[np.float32] = image_embedding.extra_features[0][None]
        backbone_fpn_2: NDArray[np.float32] = image_embedding.extra_features[1][None]
        vision_pos_enc_2: NDArray[np.float32] = image_embedding.extra_features[2][None]

        prompt: types.Prompt
        if request.prompt is None:
            prompt = types.Prompt(texts=["visual"])
            logger.warning(
                "Prompt is not given, so using 'visual' as prompt: {!r}", prompt
            )
        else:
            prompt = request.prompt

        text_prompt: str = "visual"
        if prompt.texts:
            if len(prompt.texts) > 1:
                logger.warning(
                    "Only first text prompt is used, ignoring: {}", prompt.texts[1:]
                )
            text_prompt = prompt.texts[0]

        score_threshold: float = (
            0 if prompt.score_threshold is None else prompt.score_threshold
        )

        tokens: NDArray[np.int64] = tokenize(texts=[text_prompt], context_length=32)
        outputs = self._inference_sessions["language_encoder"].run(
            None, {"tokens": tokens}
        )
        language_mask: NDArray[np.bool_] = cast(NDArray[np.bool_], outputs[0])
        language_features: NDArray[np.float32] = cast(NDArray[np.float32], outputs[1])

        box_coords: NDArray[np.float32]
        box_labels: NDArray[np.int64]
        box_masks: NDArray[np.bool_]
        if prompt.points is None and prompt.point_labels is None:
            box_coords = np.array([[[0, 0, 0, 0]]], dtype=np.float32)
            box_labels = np.array([[1]], dtype=np.int64)
            box_masks = np.array([[True]], dtype=np.bool_)  # masked out
        elif prompt.points is None or prompt.point_labels is None:
            raise ValueError(
                "both points and point_labels must be provided together: "
                f"{prompt.points=}, {prompt.point_labels=}"
            )
        else:
            if len(prompt.points) != 2 or prompt.points.shape[1] != 2:
                raise ValueError(
                    "only two point prompts (left-top, right-bottom) are supported: "
                    f"{prompt.points=}, {prompt.point_labels=}"
                )
            if prompt.point_labels.tolist() != [2, 3]:
                raise ValueError(
                    "only point labels for box prompts (2: left-top, 3: right-bottom) "
                    f"are supported: {prompt.point_labels=}"
                )
            points_normalized: NDArray[np.float32] = (
                prompt.points / [original_width, original_height]
            ).astype(np.float32)
            # (x_center, y_center, width, height)
            box_coords = np.r_[
                (points_normalized[0] + points_normalized[1]) / 2,
                points_normalized[1] - points_normalized[0],
            ][None, None]
            box_labels = np.array([[1]], dtype=np.int64)
            box_masks = np.array([[False]], dtype=np.bool_)

        outputs = self._inference_sessions["decoder"].run(
            None,
            {
                "original_height": np.array(original_height, dtype=np.int64),
                "original_width": np.array(original_width, dtype=np.int64),
                "backbone_fpn_0": backbone_fpn_0,
                "backbone_fpn_1": backbone_fpn_1,
                "backbone_fpn_2": backbone_fpn_2,
                "vision_pos_enc_2": vision_pos_enc_2,
                "language_mask": language_mask,
                "language_features": language_features,
                "box_coords": box_coords,
                "box_labels": box_labels,
                "box_masks": box_masks,
            },
        )
        boxes: NDArray[np.float32] = cast(NDArray[np.float32], outputs[0])
        scores: NDArray[np.float32] = cast(NDArray[np.float32], outputs[1])
        masks: NDArray[np.bool_] = cast(NDArray[np.bool_], outputs[2])

        annotations: list[types.Annotation] = []
        for i in range(len(scores)):
            if scores[i] < score_threshold:
                continue
            box = boxes[i]
            annotations.append(
                types.Annotation(
                    mask=masks[i, 0] > 0,
                    bounding_box=types.BoundingBox(
                        xmin=int(box[0]),
                        ymin=int(box[1]),
                        xmax=int(box[2]),
                        ymax=int(box[3]),
                    ),
                    text=text_prompt,
                    score=float(scores[i]),
                )
            )

        return types.GenerateResponse(
            model=self.name,
            image_embedding=image_embedding,
            annotations=annotations,
        )



from __future__ import annotations

import dataclasses
import os
import shutil
import urllib.parse

import gdown
from loguru import logger


@dataclasses.dataclass
class Blob:
    url: str
    hash: str
    attachments: list[Blob] = dataclasses.field(default_factory=list)

    @property
    def filename(self) -> str:
        return os.path.basename(urllib.parse.urlparse(self.url).path)

    @property
    def path(self) -> str:
        base = os.path.expanduser("~/.cache/osam/models/blobs")
        if self.attachments:
            # Windows can't use ':' for directory names
            safe_hash = self.hash.replace("sha256:", "sha256-")
            return f"{base}/{safe_hash}/{self.filename}"
        else:
            return f"{base}/{self.hash}"

    @property
    def size(self) -> int | None:
        if not os.path.exists(self.path):
            return None
        total = os.stat(self.path).st_size
        for attachment in self.attachments:
            attachment_path: str = os.path.join(
                os.path.dirname(self.path), attachment.filename
            )
            if not os.path.exists(attachment_path):
                return None
            total += os.stat(attachment_path).st_size
        return total

    @property
    def modified_at(self) -> float | None:
        if not os.path.exists(self.path):
            return None
        latest = os.stat(self.path).st_mtime
        for attachment in self.attachments:
            attachment_path: str = os.path.join(
                os.path.dirname(self.path), attachment.filename
            )
            if not os.path.exists(attachment_path):
                return None
            latest = max(latest, os.stat(attachment_path).st_mtime)
        return latest

    def pull(self):
        if self.attachments:
            blob_dir: str = os.path.dirname(self.path)
            if os.path.isfile(blob_dir):
                logger.warning("Removing file {!r} to create blob directory", blob_dir)
                os.remove(blob_dir)
            os.makedirs(blob_dir, exist_ok=True)
            gdown.cached_download(url=self.url, path=self.path, hash=self.hash)
            for attachment in self.attachments:
                attachment_path: str = os.path.join(
                    os.path.dirname(self.path), attachment.filename
                )
                gdown.cached_download(
                    url=attachment.url, path=attachment_path, hash=attachment.hash
                )
        else:
            gdown.cached_download(url=self.url, path=self.path, hash=self.hash)

    def remove(self):
        if self.attachments:
            dir_path = os.path.dirname(self.path)
            if os.path.exists(dir_path):
                logger.debug("Removing blob directory {!r}", dir_path)
                shutil.rmtree(dir_path)
            else:
                logger.warning("Blob directory {!r} not found", dir_path)
        else:
            if os.path.exists(self.path):
                logger.debug("Removing blob {!r}", self.path)
                os.remove(self.path)
            else:
                logger.warning("Blob {!r} not found", self.path)



from typing import cast

import imgviz
import numpy as np
from loguru import logger
from numpy.typing import NDArray

from ... import apis
from ... import types
from . import clip


class _YoloWorld(types.Model):
    _input_size: int

    def generate(self, request: types.GenerateRequest) -> types.GenerateResponse:
        if request.prompt is None:
            prompt = types.Prompt(texts=["anything"])
            logger.warning(
                "Prompt is not given, so using 'anything' as prompt: {prompt!r}",
                prompt=prompt,
            )
        else:
            prompt = request.prompt

        if prompt.texts is None:
            raise ValueError("prompt.texts is required: prompt=%r", prompt)
        token = clip.tokenize(texts=prompt.texts + [" "])
        outputs = self._inference_sessions["textual"].run(None, {"input": token})
        text_features: NDArray[np.float32] = cast(NDArray[np.float32], outputs[0])
        text_features = text_features / np.linalg.norm(
            text_features, ord=2, axis=1, keepdims=True
        )
        #
        if request.image is None:
            raise ValueError("request.image is required: request=%r", request)
        transformed_image, original_image_hw, padding_hw = _transform_image(
            image=request.image, input_size=self._input_size
        )
        yolo_outputs = self._inference_sessions["yolo"].run(
            output_names=["scores", "boxes"],
            input_feed={
                "images": transformed_image[None],
                "text_features": text_features[None],
            },
        )
        scores = cast(NDArray[np.float32], yolo_outputs[0])[0]
        bboxes = cast(NDArray[np.float32], yolo_outputs[1])[0]
        #
        iou_threshold = 1.0 if prompt.iou_threshold is None else prompt.iou_threshold
        score_threshold = (
            0.0 if prompt.score_threshold is None else prompt.score_threshold
        )
        max_annotations = (
            len(bboxes) if prompt.max_annotations is None else prompt.max_annotations
        )
        bboxes, scores, labels, _indices = apis.non_maximum_suppression(
            boxes=bboxes,
            scores=scores,
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            max_num_detections=max_annotations,
        )
        bboxes = _untransform_bboxes(
            bboxes=bboxes,
            input_size=self._input_size,
            original_image_hw=original_image_hw,
            padding_hw=padding_hw,
        )
        annotations = [
            types.Annotation(
                bounding_box=types.BoundingBox(
                    xmin=bbox[0], ymin=bbox[1], xmax=bbox[2], ymax=bbox[3]
                ),
                text=prompt.texts[label],
                score=score,
            )
            for bbox, label, score in zip(bboxes, labels, scores)
        ]
        return types.GenerateResponse(
            model=self.name,
            image_embedding=None,
            annotations=annotations,
        )


class YoloWorldXL(_YoloWorld):
    name: str = "yoloworld:latest"

    _input_size = 640
    _blobs = {
        "textual": types.Blob(
            url="https://clip-as-service.s3.us-east-2.amazonaws.com/models/onnx/ViT-B-32/textual.onnx",
            hash="sha256:55c85d8cbb096023781c1d13c557eb95d26034c111bd001b7360fdb7399eec68",
        ),
        "yolo": types.Blob(
            url="https://github.com/wkentaro/yolo-world-onnx/releases/download/v0.1.0/yolo_world_v2_xl_vlpan_bn_2e-3_100e_4x8gpus_obj365v1_goldg_train_lvis_minival.onnx",
            hash="sha256:92660c6456766439a2670cf19a8a258ccd3588118622a15959f39e253731c05d",
        ),
    }


def _transform_image(
    image: np.ndarray, input_size: int
) -> tuple[np.ndarray, tuple[int, int], tuple[int, int]]:
    height, width = image.shape[:2]

    scale = input_size / max(height, width)
    image_resized = imgviz.resize(
        image,
        height=int(height * scale),
        width=int(width * scale),
        interpolation="linear",
    )
    pad_height = input_size - image_resized.shape[0]
    pad_width = input_size - image_resized.shape[1]
    image_resized = np.pad(
        image_resized,
        (
            (pad_height // 2, pad_height - pad_height // 2),
            (pad_width // 2, pad_width - pad_width // 2),
            (0, 0),
        ),
        mode="constant",
        constant_values=114,
    )
    input_image = image_resized.transpose(2, 0, 1).astype(np.float32) / 255.0
    return input_image, (height, width), (pad_height, pad_width)


def _untransform_bboxes(
    bboxes: np.ndarray,
    input_size: int,
    original_image_hw: tuple[int, int],
    padding_hw: tuple[int, int],
) -> np.ndarray:
    bboxes -= np.array([padding_hw[1] // 2, padding_hw[0] // 2] * 2)
    bboxes /= input_size / max(original_image_hw)
    bboxes[:, 0::2] = np.clip(bboxes[:, 0::2], 0, original_image_hw[1])
    bboxes[:, 1::2] = np.clip(bboxes[:, 1::2], 0, original_image_hw[0])
    bboxes = bboxes.round().astype(int)
    return bboxes



import base64
from typing import Union
from typing import cast

import numpy as np
import numpy.typing as npt
import pydantic


class ImageEmbedding(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    original_height: int
    original_width: int
    embedding: npt.NDArray[np.float32]
    extra_features: list[npt.NDArray[np.float32]] = pydantic.Field(default_factory=list)

    @pydantic.field_validator("embedding", mode="before")
    def validate_embedding(
        cls, embedding: Union[npt.NDArray[np.float32], list[list[list[float]]]]
    ) -> npt.NDArray[np.float32]:
        if isinstance(embedding, list):
            embedding = np.array(embedding, dtype=np.float32)
        if embedding.ndim != 3:
            raise ValueError("embedding must be of dimension 3, not %r", embedding.ndim)
        if embedding.dtype != np.float32:
            raise ValueError(
                "embedding must be of type float, but got %r", embedding.dtype
            )
        return embedding

    @pydantic.field_serializer("embedding")
    def serialize_embedding(self, embedding: np.ndarray) -> list[list[list[float]]]:
        return embedding.tolist()

    @pydantic.field_validator("extra_features", mode="before")
    def validate_extra_features(
        cls, extra_features: list[Union[npt.NDArray[np.float32], dict]]
    ) -> list[npt.NDArray[np.float32]]:
        extra_features_ndarray: list[npt.NDArray[np.float32]] = []
        for i, high_res_feature in enumerate(extra_features):
            high_res_feature_ndarray: npt.NDArray[np.float32]
            if isinstance(high_res_feature, np.ndarray):
                high_res_feature_ndarray = cast(
                    npt.NDArray[np.float32], high_res_feature
                )
            elif isinstance(high_res_feature, dict):
                if {"data", "shape", "dtype"} != set(high_res_feature.keys()):
                    raise ValueError(
                        "extra_features must have keys {'data', 'shape', 'dtype'}, "
                        "but got %r",
                        set(high_res_feature.keys()),
                    )
                high_res_feature_ndarray = np.frombuffer(
                    base64.b64decode(high_res_feature["data"]),
                    dtype=high_res_feature["dtype"],
                ).reshape(high_res_feature["shape"])
            else:
                raise ValueError(
                    "extra_features must be either a numpy array or a dictionary, "
                    "but got %r",
                    high_res_feature,
                )
            del high_res_feature

            if high_res_feature_ndarray.ndim != 3:
                raise ValueError(
                    "extra_features must be of dimension 3, not %r",
                    high_res_feature_ndarray.ndim,
                )
            if high_res_feature_ndarray.dtype != np.float32:
                raise ValueError(
                    "extra_features must be of type float, but got %r",
                    high_res_feature_ndarray.dtype,
                )
            extra_features_ndarray.append(high_res_feature_ndarray)
        return extra_features_ndarray

    @pydantic.field_serializer("extra_features")
    def serialize_extra_features(self, extra_features: list[np.ndarray]) -> list[dict]:
        return [_ndarray_to_json(arr=arr) for arr in extra_features]


def _ndarray_to_json(arr: np.ndarray) -> dict:
    return {
        "shape": arr.shape,
        "dtype": str(arr.dtype),
        "data": base64.b64encode(arr.tobytes()).decode(),
    }



from __future__ import annotations

import abc
import hashlib
from typing import Dict
from typing import Optional
from typing import Sequence

import numpy as np
import onnxruntime
from loguru import logger

from ._blob import Blob
from ._generate import GenerateRequest
from ._generate import GenerateResponse
from ._image_embedding import ImageEmbedding


class Model(abc.ABC):
    name: str

    _blobs: Dict[str, Blob]
    _inference_sessions: Dict[str, onnxruntime.InferenceSession]

    def __init__(self):
        logger.debug("Initializing model {name}", name=self.name)
        self.pull()
        self._inference_sessions = _load_inference_sessions(blobs=self._blobs)

    @classmethod
    def pull(cls):
        for blob in cls._blobs.values():
            blob.pull()

    @classmethod
    def remove(cls):
        for blob in cls._blobs.values():
            blob.remove()

    @classmethod
    def get_id(cls) -> str:
        return hashlib.md5(
            "+".join(blob.hash for blob in cls._blobs.values()).encode()
        ).hexdigest()[:12]

    @classmethod
    def get_size(cls) -> Optional[int]:
        size = 0
        for blob in cls._blobs.values():
            if blob.size is None:
                return None
            size += blob.size
        return size

    @classmethod
    def get_modified_at(cls) -> Optional[float]:
        modified_at: float = 0
        for blob in cls._blobs.values():
            if blob.modified_at is None:
                return None
            modified_at = max(modified_at, blob.modified_at)
        return modified_at

    def encode_image(self, image: np.ndarray) -> ImageEmbedding:
        raise NotImplementedError

    @abc.abstractmethod
    def generate(self, request: GenerateRequest) -> GenerateResponse:
        pass


def _load_inference_sessions(
    blobs: dict[str, Blob],
) -> dict[str, onnxruntime.InferenceSession]:
    providers: Sequence[str] | None = None
    inference_sessions: dict[str, onnxruntime.InferenceSession] = {}
    for key, blob in blobs.items():
        inference_session: onnxruntime.InferenceSession = _load_inference_session(
            blob=blob, providers=providers
        )
        providers = inference_session.get_providers()
        inference_sessions[key] = inference_session
    logger.info(
        "Initialized inference sessions with providers {providers!r}",
        providers=providers,
    )
    return inference_sessions


def _load_inference_session(
    blob: Blob, providers: list[str] | None = None
) -> onnxruntime.InferenceSession:
    try:
        # Try to use all of the available providers e.g., cuda, tensorrt.
        if providers is None:
            if "CUDAExecutionProvider" in onnxruntime.get_available_providers():  # type: ignore[possibly-missing-attribute]
                providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]
            else:
                providers = ["CPUExecutionProvider"]
        inference_session = onnxruntime.InferenceSession(blob.path, providers=providers)
    except Exception as e:
        # Even though there is fallback in onnxruntime, it won't always work.
        # e.g., CUDA is installed and CUDA_PATH is set, but CUDA_VISIBLE_DEVICES
        # is empty. We fallback to cpu in such cases.
        logger.error(
            "Failed to create inference session with providers {providers!r}. "
            "Falling back to ['CPUExecutionProvider']: {e}",
            providers=providers,
            e=e,
        )
        providers = ["CPUExecutionProvider"]
        inference_session = onnxruntime.InferenceSession(blob.path, providers=providers)
    return inference_session



import logging
import os

import fastapi  # type: ignore[unresolved-import]

from . import apis
from . import types

app: fastapi.FastAPI = fastapi.FastAPI()


@app.get("/")
async def index():
    return fastapi.Response("osam is running")


@app.post("/api/generate")
async def generate(request: types.GenerateRequest) -> types.GenerateResponse:
    try:
        return apis.generate(request=request)
    except ValueError as e:
        raise fastapi.HTTPException(400, str(e))


# https://github.com/tiangolo/fastapi/issues/1508#issuecomment-638365277
@app.on_event("startup")
async def startup_event():
    uvicorn_logger = logging.getLogger("uvicorn")
    handler = logging.FileHandler(os.path.expanduser("~/.cache/osam/osam.log"))
    handler.setLevel(logging.DEBUG)
    fmt = "%(asctime)s [%(levelname)s] %(module)s:%(funcName)s:%(lineno)s - %(message)s"
    handler.setFormatter(logging.Formatter(fmt=fmt))
    uvicorn_logger.addHandler(handler)



from .efficientsam import EfficientSam10m  # noqa: F401
from .efficientsam import EfficientSam30m  # noqa: F401
from .sam import Sam100m  # noqa: F401
from .sam import Sam300m  # noqa: F401
from .sam import Sam600m  # noqa: F401
from .sam2 import Sam2BasePlus  # noqa: F401
from .sam2 import Sam2Large  # noqa: F401
from .sam2 import Sam2Small  # noqa: F401
from .sam2 import Sam2Tiny  # noqa: F401
from .sam3 import Sam3  # noqa: F401
from .yoloworld import YoloWorldXL  # noqa: F401



#!/usr/bin/env python3

import argparse
from typing import Type

import onnx
from loguru import logger

import osam


def count_parameters(onnx_file) -> int:
    model = onnx.load(onnx_file)
    total_params = 0

    for initializer in model.graph.initializer:
        param_count = 1
        for dim in initializer.dims:
            param_count *= dim
        total_params += param_count

    return total_params


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "model",
        type=str,
        choices=[model_type.name for model_type in osam.apis.registered_model_types],
    )
    args = parser.parse_args()

    model_type: Type[osam.types.Model] = osam.apis.get_model_type_by_name(
        name=args.model
    )
    model_type.pull()

    blob: osam.types.Blob
    num_params: int = 0
    for blob in model_type._blobs.values():
        logger.debug(f"Counting parameters in {blob.path}")
        num_params += count_parameters(onnx_file=blob.path)

    logger.debug(
        "Number of parameters: "
        f"{num_params} = {num_params / 1e6:.2f}M = {num_params / 1e9:.2f}B"
    )


if __name__ == "__main__":
    main()



# https://github.com/openai/CLIP/blob/main/clip/clip.py

import numpy as np

from .simple_tokenizer import SimpleTokenizer

_tokenizer = SimpleTokenizer()


def tokenize(
    texts: list[str], context_length: int = 77, truncate: bool = False
) -> np.ndarray:
    sot_token = _tokenizer.encoder["<|startoftext|>"]
    eot_token = _tokenizer.encoder["<|endoftext|>"]
    all_tokens = [[sot_token] + _tokenizer.encode(text) + [eot_token] for text in texts]

    result = np.zeros((len(all_tokens), context_length), dtype=np.int64)

    for i, tokens in enumerate(all_tokens):
        if len(tokens) > context_length:
            if truncate:
                tokens = tokens[:context_length]
                tokens[-1] = eot_token
            else:
                raise RuntimeError(
                    f"Input {texts[i]} is too long for context length {context_length}"
                )
        result[i, : len(tokens)] = tokens

    return result



from typing import cast

import numpy as np
import numpy.typing as npt
import onnxruntime

from osam import types

from . import _images


def generate_mask_from_image_embedding(
    decoder_session: onnxruntime.InferenceSession,
    image_embedding: types.ImageEmbedding,
    prompt: types.Prompt,
    input_size: int,
) -> npt.NDArray[np.bool_]:
    if prompt.points is None or prompt.point_labels is None:
        raise ValueError("Prompt must contain points and point_labels: %r" % prompt)

    onnx_coord: npt.NDArray[np.float32] = np.concatenate(
        [prompt.points, np.array([[0.0, 0.0]])], axis=0
    )[None, :, :]
    onnx_label: npt.NDArray[np.float32] = np.concatenate(
        [prompt.point_labels, np.array([-1])], axis=0
    )[None, :].astype(np.float32)

    scale: float
    new_height: int
    new_width: int
    scale, new_height, new_width = _images.compute_scale_to_resize_image(
        height=image_embedding.original_height,
        width=image_embedding.original_width,
        target_size=input_size,
    )
    onnx_coord = (
        onnx_coord.astype(float)
        * (
            new_width / image_embedding.original_width,
            new_height / image_embedding.original_height,
        )
    ).astype(np.float32)

    onnx_mask_input: npt.NDArray[np.float32] = np.zeros(
        (1, 1, 256, 256), dtype=np.float32
    )
    onnx_has_mask_input: npt.NDArray[np.float32] = np.array([-1], dtype=np.float32)

    decoder_inputs: dict[str, npt.NDArray] = {
        "image_embeddings": image_embedding.embedding[None, :, :, :],
        "point_coords": onnx_coord,
        "point_labels": onnx_label,
        "mask_input": onnx_mask_input,
        "has_mask_input": onnx_has_mask_input,
        "orig_im_size": np.array(
            (image_embedding.original_height, image_embedding.original_width),
            dtype=np.float32,
        ),
    }

    masks, _, _ = decoder_session.run(None, decoder_inputs)
    masks = cast(npt.NDArray[np.bool_], masks)
    mask: npt.NDArray[np.bool_] = masks[0, 0] > 0.0  # (1, 1, H, W) -> (H, W)
    return mask



from typing import cast

import numpy as np
import numpy.typing as npt
import onnxruntime

from osam import types


def generate_mask_from_image_embedding(
    decoder_session: onnxruntime.InferenceSession,
    image_embedding: types.ImageEmbedding,
    prompt: types.Prompt,
) -> npt.NDArray[np.bool_]:
    input_point: npt.NDArray[np.float32] = np.array(prompt.points, dtype=np.float32)
    input_label: npt.NDArray[np.float32] = np.array(
        prompt.point_labels, dtype=np.float32
    )

    # batch_size, embedding_dim, height, width
    batched_image_embedding = image_embedding.embedding[None, :, :, :]
    # batch_size, num_queries, num_points, 2
    batched_point_coords = input_point[None, None, :, :]
    # batch_size, num_queries, num_points
    batched_point_labels = input_label[None, None, :]

    decoder_inputs = {
        "image_embeddings": batched_image_embedding,
        "batched_point_coords": batched_point_coords,
        "batched_point_labels": batched_point_labels,
        "orig_im_size": np.array(
            (image_embedding.original_height, image_embedding.original_width),
            dtype=np.int64,
        ),
    }

    masks, _, _ = decoder_session.run(None, decoder_inputs)
    masks = cast(npt.NDArray[np.bool_], masks)
    mask: npt.NDArray[np.bool_] = masks[0, 0, 0, :, :] > 0  # (1, 1, 3, H, W) -> (H, W)

    return mask



import pathlib

import imgviz
import numpy as np
import pytest

from . import apis
from . import types

here = pathlib.Path(__file__).resolve().parent


@pytest.mark.parametrize(
    "model",
    [
        "efficientsam:10m",
        pytest.param("efficientsam:latest", marks=pytest.mark.extra),
        pytest.param("sam:100m", marks=pytest.mark.extra),
        pytest.param("sam:300m", marks=pytest.mark.extra),
        pytest.param("sam:latest", marks=pytest.mark.extra),
        pytest.param("sam2:tiny", marks=pytest.mark.extra),
        pytest.param("sam2:small", marks=pytest.mark.extra),
        pytest.param("sam2:latest", marks=pytest.mark.extra),
        pytest.param("sam2:large", marks=pytest.mark.extra),
    ],
)
def test_generate_point_to_mask(model: str) -> None:
    image = imgviz.io.imread(here / "_data" / "dogs.jpg")
    request: types.GenerateRequest = types.GenerateRequest(model=model, image=image)
    response: types.GenerateResponse = apis.generate(request=request)

    assert response.model == model

    assert len(response.annotations) == 1
    annotation: types.Annotation = response.annotations[0]
    assert annotation.text is None
    assert annotation.score is None
    assert annotation.mask is not None
    assert annotation.mask.dtype == bool
    assert annotation.mask.shape == image.shape[:2]
    assert annotation.bounding_box is not None


@pytest.mark.parametrize(
    "model, has_mask",
    [
        ("sam3:latest", True),
        pytest.param("yoloworld:latest", False, marks=pytest.mark.extra),
    ],
)
def test_generate_text_to_bounding_box(model: str, has_mask: bool) -> None:
    image = imgviz.io.imread(here / "_data" / "dogs.jpg")
    request: types.GenerateRequest = types.GenerateRequest(
        model=model, image=image, prompt=types.Prompt(texts=["dog"])
    )
    response: types.GenerateResponse = apis.generate(request=request)

    assert response.model == model

    assert len(response.annotations) == 3
    for annotation in response.annotations:
        assert annotation.bounding_box is not None
        assert annotation.text == "dog"
        assert isinstance(annotation.score, float)
        if has_mask:
            assert annotation.mask is not None
            assert annotation.mask.dtype == bool
            assert annotation.mask.shape == image.shape[:2]
        else:
            assert annotation.mask is None


@pytest.mark.parametrize("model", ["sam2:tiny"])
def test_generate_box_to_mask_sam2(model: str) -> None:
    image = imgviz.io.imread(here / "_data" / "dogs.jpg")
    request: types.GenerateRequest = types.GenerateRequest(
        model=model,
        image=image,
        prompt=types.Prompt(
            points=np.array([[1233, 376], [1649, 691]], dtype=np.float32),
            point_labels=np.array([2, 3], dtype=np.int32),
        ),
    )
    response: types.GenerateResponse = apis.generate(request=request)

    assert response.model == model

    assert len(response.annotations) == 1
    for annotation in response.annotations:
        assert annotation.text is None
        assert annotation.score is None
        assert annotation.mask is not None
        assert annotation.mask.dtype == bool
        assert annotation.mask.shape == image.shape[:2]
        assert annotation.bounding_box is not None


@pytest.mark.parametrize("model", ["sam3:latest"])
def test_generate_box_to_mask_sam3(model: str) -> None:
    image = imgviz.io.imread(here / "_data" / "dogs.jpg")
    request: types.GenerateRequest = types.GenerateRequest(
        model=model,
        image=image,
        prompt=types.Prompt(
            points=np.array([[1233, 376], [1649, 691]], dtype=np.float32),
            point_labels=np.array([2, 3], dtype=np.int32),
        ),
    )
    response: types.GenerateResponse = apis.generate(request=request)

    assert response.model == model

    # SAM3 returns multiple mask candidates (unlike SAM2 which returns 1)
    assert len(response.annotations) == 3
    for annotation in response.annotations:
        assert annotation.text == "visual"
        assert annotation.score is not None
        assert annotation.mask is not None
        assert annotation.mask.dtype == bool
        assert annotation.mask.shape == image.shape[:2]
        assert annotation.bounding_box is not None



import numpy as np
import numpy.typing as npt
import PIL.Image


def compute_scale_to_resize_image(
    height: int, width: int, target_size: int
) -> tuple[float, int, int]:
    scale: float
    new_height: int
    new_width: int
    if width > height:
        scale = target_size / width
        new_height = int(round(height * scale))
        new_width = target_size
    else:
        scale = target_size / height
        new_height = target_size
        new_width = int(round(width * scale))
    return scale, new_height, new_width


def resize_image(image: npt.NDArray, target_size: int) -> tuple[float, npt.NDArray]:
    scale: float
    new_height: int
    new_width: int
    scale, new_height, new_width = compute_scale_to_resize_image(
        height=image.shape[0],
        width=image.shape[1],
        target_size=target_size,
    )
    scaled_image: npt.NDArray = np.asarray(
        PIL.Image.fromarray(image).resize(
            (new_width, new_height),
            resample=PIL.Image.BILINEAR,  # type: ignore[attr-defined]
        )
    )
    return scale, scaled_image
